/*
  PidgiMPD
  Copyright (C) 2006 Erik Karlsson

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#define PURPLE_PLUGINS 1

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <string.h>
#include <inttypes.h>
#include <inttypes.h>

#include <plugin.h>

#include <cmds.h>
#include <conversation.h>
#include <debug.h>
#include <prefs.h>
#include <version.h>
#include <signals.h>
#include <notify.h>
#include <status.h>
#include <core.h>

#include <gtkconv.h>
#include <gtkplugin.h>
#include <gtkimhtml.h>
#include <gtkprefs.h>
#include <gtkutils.h>

#include <libmpd/libmpd.h>

#include <math.h>

enum CommandID
{
	COMMAND_PLAY = 0,
	COMMAND_PAUSE,
	COMMAND_STOP,
	COMMAND_NEXT,
	COMMAND_PREV,
	COMMAND_REPEAT,
	COMMAND_RANDOM,
	COMMAND_SEND,
	COMMAND_STATUS,
	COMMAND_HELP
};

/* Globals */
static MpdObj      *mo            = NULL;
static gchar       *format_syntax = NULL,
                   *help          = NULL,
                   *command_help[COMMAND_HELP + 1];
static PurpleCmdId pidgimpd_cmd;
static guint       pidgimpd_pref_change,
                   pidgimpd_update_status;

/* Function declarations */
static void pidgimpd_status_changed_cb(
    MpdObj            *mo,
    ChangedStatusType what,
    void              *userdata);
static void pidgimpd_connection_changed_cb(
    MpdObj *mo,
    int    connect,
    void   *userdata);


static PurplePluginPrefFrame * get_plugin_pref_frame(PurplePlugin *plugin)
{
	PurplePluginPrefFrame *frame = NULL;
	PurplePluginPref      *ppref = NULL;
 
	frame = purple_plugin_pref_frame_new();
 
	ppref = purple_plugin_pref_new_with_label("PidgiMPD settings");
	purple_plugin_pref_frame_add(frame, ppref);
 
	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/hostname",
		"Host to connect to:");
	purple_plugin_pref_set_max_length(ppref, 128);
	purple_plugin_pref_frame_add(frame, ppref);
 
	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/port",
		"Server port:");
	purple_plugin_pref_set_bounds(ppref, 1024, 65535);
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/password",
		"Password (optional):");
	purple_plugin_pref_set_max_length(ppref, 128);
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/title_on_away",
		"Set away message to song title");
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/showformat",
		"Away message format:");
	purple_plugin_pref_set_max_length(ppref, 128);
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/send_on_new",
		"Send song title to buddies in active conversations on title change");
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/show_on_new",
		"Display song title in conversations on title change");
	purple_plugin_pref_frame_add(frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label(
		"/plugins/gtk/audio/pidgimpd/songformat",
		"Conversation format:");
	purple_plugin_pref_set_max_length(ppref, 128);
	purple_plugin_pref_frame_add(frame, ppref);

	return frame;
}

static void pidgimpd_send_text(PurpleConversation *c, const gchar *text)
{
	PurpleConversationType type;

	g_return_if_fail(c);

	if(!text)
		return;
	
	type = purple_conversation_get_type(c);

	switch(type)
	{
		case PURPLE_CONV_TYPE_IM:
			purple_conv_im_send(PURPLE_CONV_IM(c), text);
			break;
		case PURPLE_CONV_TYPE_CHAT:
				purple_conv_chat_send_with_flags(
				    PURPLE_CONV_CHAT(c),
				    text,
				    PURPLE_MESSAGE_RAW);
			break;
		case PURPLE_CONV_TYPE_UNKNOWN:
		case PURPLE_CONV_TYPE_MISC:
		default:
			break;
	}
}

static void pidgimpd_display_text(PurpleConversation *c, const gchar *text)
{
	PurpleConversationType type;

	g_return_if_fail(c);

	type = purple_conversation_get_type(c);

	if(!text)
		return;

	switch(type)
	{
		case PURPLE_CONV_TYPE_IM:
			purple_conv_im_write(
			    PURPLE_CONV_IM(c),
			    NULL,
			    text,
			    PURPLE_MESSAGE_NO_LOG,
			    time(NULL));
			break;
		case PURPLE_CONV_TYPE_CHAT:
			purple_conv_chat_write(
			    PURPLE_CONV_CHAT(c),
			    NULL,
			    text,
			    PURPLE_MESSAGE_NO_LOG,
			    time(NULL));
			break;
		case PURPLE_CONV_TYPE_UNKNOWN:
		case PURPLE_CONV_TYPE_MISC:
		default:
			break;
	}
}

static const gchar * pidgimpd_status_string()
{
	if(!mpd_check_connected(mo))
	{
		return "not connected";
	}
	else
		switch(mpd_player_get_state(mo))
		{
			case MPD_PLAYER_PAUSE:
				return "paused";
			case MPD_PLAYER_PLAY:
				return "playing";
			case MPD_PLAYER_STOP:
				return "stopped";
			case MPD_PLAYER_UNKNOWN:
			default:
				return "unknown status";
		}
}

static const gchar * pidgimpd_guint_to_timestring(guint time)
{
	GString *str = NULL;
	gchar   *ret = NULL;

	str = g_string_new("");

	g_string_printf(
		str,
		"%02d:%02d",
		time/60,
		time%60);

	ret = g_string_free(str, FALSE);

	return ret;
}

static const gchar * pidgimpd_song_info(int song_id, const gchar *songformat)
{
	GString      *str = NULL;
	gchar        *ret = NULL;
	mpd_Song     *song = NULL;
	unsigned int i,
	             formatlength,
	             time;
	gboolean     percent;
	
	if(!mpd_check_connected(mo))
		return NULL;

	/* In case a command was executed before, we need to get the new data */
	mpd_status_update(mo);
	
	if((formatlength = strlen(songformat)) == 0)
	{
		/*  Retrieve the formatting of the string */
		songformat = (gchar*)purple_prefs_get_string(
		    "/plugins/gtk/audio/pidgimpd/songformat");
		formatlength = strlen(songformat);
	}
	
	/*  Grab song struct from MPD object */
	if(song_id == -1)
		song = mpd_playlist_get_current_song(mo);
	else
		song = mpd_playlist_get_song(mo, song_id);

	if(!song)
		return NULL;

	str = g_string_new(NULL);

	percent = FALSE;

	for(i = 0; i < formatlength; i++)
	{
		if(songformat[i] == '%')
		{
			if(percent)
			{
				percent = FALSE;
			}
			else
			{
				percent = TRUE;
			
				continue;
			}
		}

		if(!percent)
		{
			g_string_append_c(str, songformat[i]);

			continue;
		}
		
		switch(songformat[i])
		{
			case 'a':
				if(song->artist != NULL)
					g_string_append(str, song->artist);
				break;
			case 't':
				if(song->title != NULL)
					g_string_append(str, song->title);
				break;
			case 'A':
				if(song->album != NULL)
					g_string_append(str, song->album);
				break;
			case 'f':
				if(song->file != NULL)
					g_string_append(str, song->file);
				break;
			case 'T':
				if(song->track != NULL)
					g_string_append(str, song->track+1);
				break;
			case 'g':
				if(song->genre != NULL)
					g_string_append(str, song->genre);
				break;
			case 'c':
				if(song->composer != NULL)
					g_string_append(str, song->composer);
				break;
			case 'i':
				g_string_append_printf(
				    str,
				    "%d",
				    mpd_player_get_current_song_id(mo)+1);
			case 'n':
				g_string_append_printf(str, "%d", mpd_stats_get_total_songs(mo));
				break;
			case 's':
				g_string_append(str, pidgimpd_status_string());
				break;
			case 'r':
				time = song->time - mpd_status_get_elapsed_song_time(mo);

				g_string_append(str, pidgimpd_guint_to_timestring(time));
				break;
			case 'p':
				time = mpd_status_get_elapsed_song_time(mo);

				g_string_append(str, pidgimpd_guint_to_timestring(time));
				break;
			case 'l':
				time = song->time;
				g_string_append(str, pidgimpd_guint_to_timestring(time));
				break;
			default:
				break;
		}

		percent = FALSE;
	}

	ret = g_string_free(str, FALSE);

	return ret;
}

/**
 * Stolen from musictracker 0.4.1
 */
gboolean pidgimpd_status_supports_attr(PurpleStatus *status, const char *id)
{
	PurpleStatusType *status_type = NULL;
	GList            *attrs       = NULL;
	PurpleStatusAttr *attr        = NULL;

	if(status == NULL || id == NULL)
		return FALSE;

	status_type = purple_status_get_type(status);

	if(status_type == NULL)
		return FALSE;

	attrs = purple_status_type_get_attrs(status_type);
	while(attrs != NULL)
	{
		attr = (PurpleStatusAttr*)attrs->data;
		if(attr != NULL &&
		   strncasecmp(id, purple_status_attr_get_id(attr), strlen(id)) == 0)
		{
			return TRUE;
		}

		attrs = attrs->next;
	}

	return FALSE;
}

/* Stolen from libpurple status.c ;) */
static void pidgimpd_status_set_active_with_attrs(PurpleStatus *status, ...)
{
	va_list args;
	va_start(args, status);
	purple_status_set_active_with_attrs(status, TRUE, args);
	va_end(args);
}

static void pidgimpd_connection_set_away(PurpleConnection *gc,
                                         const gchar *format)
{
	PurpleAccount  *account     = NULL;
	PurplePresence *presence    = NULL;
	PurpleStatus   *status      = NULL;
	const gchar    *id          = NULL;

	/*  Not connected */
	if(purple_connection_get_state(gc) != PURPLE_CONNECTED)
	{
		return;
	}

	account = purple_connection_get_account(gc);
	/* Not active */
	if(!purple_account_get_enabled(account, purple_core_get_ui()))
	{
		return;
	}

	presence = purple_account_get_presence(account);

	status = purple_presence_get_status(presence, 
	    purple_primitive_get_id_from_type(PURPLE_STATUS_TUNE));
	if(status == NULL)
	{
		purple_debug_misc("pidgimpd", "Could not get status information " \
		                              "regarding tune from account.\n");
	}
	else
	{
		const gchar *title  = NULL,
		            *artist = NULL,
		            *album  = NULL;

		if(!format)
		{
			purple_debug_misc("pidgimpd", "Removing tune status.\n");
			purple_status_set_active(status, FALSE);
		}
		else
		{
			title  = pidgimpd_song_info(-1, "%t");
			artist = pidgimpd_song_info(-1, "%a");
			album  = pidgimpd_song_info(-1, "%A");

			purple_debug_misc("pidgimpd", "Setting tune: (%s, %s)\n",
			                  artist,
			                  title);

			pidgimpd_status_set_active_with_attrs(status,
				PURPLE_TUNE_TITLE, title,
				PURPLE_TUNE_ARTIST, artist,
				PURPLE_TUNE_ALBUM, album,
				NULL);
		}

		return;
	}

	status = purple_presence_get_active_status(presence);
	if(status == NULL)
	{
		purple_debug_misc("pidgimpd", "Could not get status information from" \
		                              " account.\n");

		return;
	}

	id = purple_status_get_id(status);
	if(pidgimpd_status_supports_attr(status, "message"))
	{
		const gchar *message = NULL;
		if(format)
			message = pidgimpd_song_info(-1, format);

		/*  Set status message, finally.. */
		purple_account_set_status(account, id, TRUE, "message", message, NULL);

/*		purple_status_set_attr_string(status, "message", song_info);
		prpl_info->set_status(account, status);
*/
	}
}

static void pidgimpd_status_set_away(const gchar *format)
{
	PurpleConnection *gc          = NULL;
	GList            *connections = NULL;

	connections = purple_connections_get_all();
	
	for(; connections != NULL; connections = connections->next)
	{
		gc = (PurpleConnection *)connections->data;

		pidgimpd_connection_set_away(gc, format);
	}
}

static void pidgimpd_status_changed_cb(
    MpdObj            *mo,
    ChangedStatusType what,
    void              *userdata)
{
	PurpleConversation *conv   = NULL;
	const char         *format = NULL;
	gboolean           send_on_new,
	                   show_on_new,
	                   title_on_away;
	int                status;
	
	format = (gchar*)purple_prefs_get_string(
	    "/plugins/gtk/audio/pidgimpd/showformat");
	send_on_new = 
	    purple_prefs_get_bool("/plugins/gtk/audio/pidgimpd/send_on_new");
	show_on_new = 
	    purple_prefs_get_bool("/plugins/gtk/audio/pidgimpd/show_on_new");
	title_on_away = 
	    purple_prefs_get_bool("/plugins/gtk/audio/pidgimpd/title_on_away");

	status = mpd_player_get_state(mo);
	if(status == MPD_PLAYER_STOP)
	{
		/* This will remove messages and tune status */
		pidgimpd_status_set_away(NULL);

		return;
	}

	/* If the new state is that a song has changed or the play state has
	 * changed, change away message and/or tune */
	if(title_on_away)
	{
		if(what & MPD_CST_SONGID ||
	       what & MPD_CST_STATE)
		{
			if(status == MPD_PLAYER_PAUSE)
			{
				gchar *temp_format = g_strconcat(format, " (%s)", NULL);

				pidgimpd_status_set_away(temp_format);

				g_free(temp_format);

				return;
			}
			
			/* Finally... */
			pidgimpd_status_set_away(format);
		}
	}

	if((send_on_new || show_on_new) &&
	   what & MPD_CST_SONGID)
	{
		GList *convs = NULL;
		const gchar *song_info = NULL,
		            *show_info = NULL;

		song_info = pidgimpd_song_info(-1, "");
		show_info = pidgimpd_song_info(-1, format);

		convs = purple_get_conversations();
		
		for(; convs != NULL; convs = convs->next)
		{
			conv = (PurpleConversation*)convs->data;
			
			if(send_on_new)
				pidgimpd_send_text(conv, song_info);
			if(show_on_new)
				pidgimpd_display_text(conv, show_info);
		}
	}
	
	mpd_status_update(mo);
}

static gboolean pidgimpd_connect(PurpleConversation *c)
{
	gchar    *hostname = NULL,
	         *password = NULL;
	uint16_t port;
	int32_t  status;
	gboolean pref_changed;
	
	hostname = (gchar*)purple_prefs_get_string(
	    "/plugins/gtk/audio/pidgimpd/hostname");
	port = purple_prefs_get_int("/plugins/gtk/audio/pidgimpd/port");
	password = (gchar*)purple_prefs_get_string(
	    "/plugins/gtk/audio/pidgimpd/password");
	pref_changed = purple_prefs_get_bool(
	    "/plugins/gtk/audio/pidgimpd/pref_changed");
	
	/*  Restore pref_changed to it's normal state */
	purple_prefs_set_bool("/plugins/gtk/audio/pidgimpd/pref_changed", FALSE);

	/*  An MPD object has not been established */
	if(mo == NULL)
		mo = mpd_new(hostname, port, password);

	mpd_set_connection_timeout(mo, 0.5);

	/*  If we are connected and  */
	if(mpd_check_connected(mo) && !pref_changed)
	{
		return TRUE;
	}
	else
	{
		mpd_set_hostname(mo, hostname);
		mpd_set_port(mo, port);
	}

	if(strlen(password) != 0)
	{
		purple_debug_misc("pidgimpd",
		                  "Attempting to connect with password.\n");

		mpd_set_password(mo, password);
		
		status = mpd_send_password(mo);
	}
	else
	{
		purple_debug_misc("pidgimpd",
		                  "Attempting to connect without password.\n");
		
		status = mpd_connect(mo);
	}

	switch(status)
	{
		case MPD_OK:
			mpd_signal_connect_status_changed(
			    mo,
			    pidgimpd_status_changed_cb,
			    NULL);
			mpd_signal_connect_connection_changed(
			    mo,
			    pidgimpd_connection_changed_cb,
			    NULL);

			return TRUE;
		case MPD_ARGS_ERROR:
		case MPD_SERVER_ERROR:
		case MPD_FATAL_ERROR:
		case MPD_STATUS_FAILED:
		case MPD_NOT_CONNECTED:
		case MPD_LOCK_FAILED:
			if(c != NULL) // Command not initiated by a user command
				pidgimpd_display_text(
				    c,
				    "PidgiMPD: An error occured while connecting to MPD, " \
				        "please check your settings.");
		default:
			purple_debug_misc("pidgimpd",
			    "An unknown error occured while connecting to MPD.\n");
			
			/*  Free object, since we are not connected. */
			mpd_free(mo);

			mo = NULL;
			break;
	}

	return FALSE;
}

static void pidgimpd_connection_changed_cb(
    MpdObj *mo,
    int    connect,
    void   *userdata)
{
}

static const gchar *pidgimpd_gchar_capitalize_first(gchar *buffer)
{
	if(strlen(buffer) > 0)
	{
		gchar c = buffer[0];
		buffer[0] = g_ascii_toupper(c);
	}

	return buffer;
}

static const gchar * pidgimpd_format_status()
{
	GString      *str  = NULL;
	gchar        *ret  = NULL;
	mpd_Song     *song = NULL;
	unsigned int time,
	             progress_int;
	float        progress_float;

	if(!mpd_check_connected(mo))
		return NULL;
	
	str = g_string_new("<br />");
	
	int status=mpd_player_get_state(mo);
	if(status == MPD_PLAYER_PLAY || status == MPD_PLAYER_PAUSE)
	{
		song = mpd_playlist_get_current_song(mo);
		
		time = mpd_status_get_elapsed_song_time(mo);
		
		progress_float = (float)mpd_status_get_elapsed_song_time(mo) /
		                 (float)(song->time);
		progress_int = (unsigned int)ceil(progress_float * 100);
		
		g_string_append_printf(str, "%s<br />[%s] #%d/%d  %s (%d%%)", 
			pidgimpd_song_info(-1, "%a - %t"),
			pidgimpd_status_string(),
			mpd_player_get_current_song_id(mo)+1,
			mpd_stats_get_total_songs(mo),
			pidgimpd_guint_to_timestring(time),
			progress_int);
	}
	
	g_string_append_printf(str, "volume: %d%%  repeat: %s  random: %s",
		mpd_status_get_volume(mo),
		(mpd_player_get_repeat(mo)?"on":"off"),
		(mpd_player_get_random(mo)?"on":"off"));

	ret = g_string_free(str, FALSE);
	
	return ret;
}

static const gchar * pidgimpd_get_command_help(const gchar *command)
{
	// TODO: Use a hash? :)
	if(strcmp(command, "play") == 0)
		return command_help[COMMAND_PLAY];
	else if(strcmp(command, "pause") == 0)
		return command_help[COMMAND_PAUSE];
	else if(strcmp(command, "stop") == 0)
		return command_help[COMMAND_STOP];
	else if(strcmp(command, "next") == 0)
		return command_help[COMMAND_NEXT];
	else if(strcmp(command, "prev") == 0)
		return command_help[COMMAND_PREV];
	else if(strcmp(command, "repeat") == 0)
		return command_help[COMMAND_REPEAT];
	else if(strcmp(command, "random") == 0)
		return command_help[COMMAND_RANDOM];
	else if(strcmp(command, "send") == 0)
		return command_help[COMMAND_SEND];
	else if(strcmp(command, "status") == 0)
		return command_help[COMMAND_STATUS];
	else if(strcmp(command, "help") == 0)
		return command_help[COMMAND_HELP];

	return NULL;
}

static PurpleCmdRet pidgimpd_do_cmd_cb(
    PurpleConversation *c,
    const gchar        *cmd,
    gchar              *args[],
    gchar              *error[],
    void               *data)
{
	gchar *format   = NULL,
	      *command  = NULL,
	      *argument = NULL;

	/*  Missing arguments, assume status command */
	if(*args == NULL && args[0] == NULL)
	{
		args[0] = g_strdup("status"); // Evil code!
	}

	command = g_ascii_strdown(args[0], strlen(args[0]));
	if(args[1] != NULL)
		argument = g_ascii_strdown(args[1], strlen(args[1]));

	/*  Connect/reconnect */
	if(!pidgimpd_connect(c))
	{
		*error = g_strdup("");
		return PURPLE_CMD_RET_FAILED;
	}

	/* This is to bring the MPD object up to speed with reality, since there's
	 * no subscription in MPD yet. */ 
	mpd_status_update(mo);

	format = (gchar*)purple_prefs_get_string(
	    "/plugins/gtk/audio/pidgimpd/showformat");

	if(strcmp(command, "play") == 0)
	{
		if(mpd_player_get_state(mo) == MPD_PLAYER_PLAY)
			pidgimpd_display_text(c, "PidgiMPD: Already playing.");
		else
		{
			mpd_player_play(mo);
			pidgimpd_display_text(c, pidgimpd_format_status());
		}
	}
	else if(!strcmp(command, "pause"))
	{
		if(mpd_player_get_state(mo) == MPD_PLAYER_PAUSE)
			pidgimpd_display_text(c, "PidgiMPD: Already paused.");
		else
		{
			mpd_player_pause(mo);
			pidgimpd_display_text(c, pidgimpd_format_status());
		}
	}
	else if(!strcmp(command, "stop"))
	{
		if(mpd_player_get_state(mo)==MPD_PLAYER_STOP)
			pidgimpd_display_text(c, "PidgiMPD: Already stopped.");
		else
		{
			mpd_player_stop(mo);
			pidgimpd_display_text(c, pidgimpd_format_status());
		}
	}
	else if(!strcmp(command, "next"))
	{
		mpd_player_next(mo);
		pidgimpd_display_text(c, pidgimpd_song_info(-1, format));
	}
	else if(!strcmp(command, "prev"))
	{
		mpd_player_prev(mo);
		pidgimpd_display_text(c, pidgimpd_song_info(-1, format));
	}
	else if(!strcmp(command, "send"))
	{
		if(mpd_player_get_state(mo) == MPD_PLAYER_PLAY)
			pidgimpd_send_text(c, pidgimpd_song_info(-1, ""));
		else
			pidgimpd_display_text(
			    c,
			    "PidgiMPD: Not currently playing any songs.");
	}
	else if(!strcmp(command, "repeat"))
	{
		if(mpd_player_get_repeat(mo))
		{
			mpd_player_set_repeat(mo, 0);
			pidgimpd_display_text(c, "PidgiMPD: Repeat: Off");
		}
		else
		{
			mpd_player_set_repeat(mo, 1);
			pidgimpd_display_text(c, "PidgiMPD: Repeat: On");
		}
	}
	else if(!strcmp(command, "random"))
	{
		if(mpd_player_get_random(mo))
		{
			mpd_player_set_random(mo, 0);
			pidgimpd_display_text(c, "PidgiMPD: Random: Off");
		}
		else
		{
			mpd_player_set_random(mo, 1);
			pidgimpd_display_text(c, "PidgiMPD: Random: On");
		}
	}
	else if(!strcmp(command, "status"))
	{
		pidgimpd_display_text(c, pidgimpd_format_status());
	}
	else if(!strcmp(command, "help"))
	{
		if(argument == NULL)
		{
			pidgimpd_display_text(c, help);
		}
		else
		{
			const gchar *helptext = pidgimpd_get_command_help(argument);

			if(helptext != NULL)
				pidgimpd_display_text(c,
				    g_strdup_printf("%s: %s", argument, helptext));
			else
				pidgimpd_display_text(c, 
					g_strdup_printf("PidgiMPD: Unknown command: %s", argument));
		}
	}
	else
	{
		*error = g_strdup_printf("PidgiMPD: Unknown command: %s", command);
		return PURPLE_CMD_RET_FAILED;
	}

	return PURPLE_CMD_RET_OK;
}

static void pidgimpd_pref_change_cb(
    const gchar    *name,
    PurplePrefType type,
    gconstpointer  val,
    gpointer       data)
{
	if(!strcmp(name, "/plugins/gtk/audio/pidgimpd/hostname") ||
	   !strcmp(name, "/plugins/gtk/audio/pidgimpd/port") ||
	   !strcmp(name, "/plugins/gtk/audio/pidgimpd/password"))
	{
		/* Change pref_changed, so that PidgiMPD will reconnect as soon as 
		 * possible. */
		purple_prefs_set_bool("/plugins/gtk/audio/pidgimpd/pref_changed", TRUE);
	}
}

static gboolean pidgimpd_update_status_cb()
{
	if(!mpd_check_connected(mo))
	{
		pidgimpd_connect(NULL);
	
		/* Update away message and such, which would be stale otherwise */
		pidgimpd_status_changed_cb(mo, MPD_CST_SONGID, NULL);
	}
	
	mpd_status_update(mo);
	
	return TRUE;
}

static gboolean plugin_load(PurplePlugin *plugin)
{
	purple_debug_misc("pidgimpd", "Loading PidgiMPD.\n");
	
	purple_prefs_set_bool("/plugins/gtk/audio/pidgimpd/pref_changed", FALSE);

	format_syntax = g_strdup(
		"%a - artist\n" \
		"%t - title\n" \
		"%A - album\n" \
		"%f - filename\n" \
		"%T - track number\n" \
		"%g - genre\n" \
		"%c - composer\n" \
		"%s - player status\n" \
		"%r - time remaining of song\n" \
		"%p - time played of song\n" \
		"%l - length of song\n" );

	help = g_strdup(
		"<pre>" \
		"mpc [&lt;play&gt; | &lt;pause&gt; | &lt;stop&gt; | &lt;next&gt; | " \
		"&lt;prev&gt; | &lt;repeat&gt; | &lt;random&gt; | &lt;send&gt; | " \
		"&lt;status&gt; | &lt;help [command]&gt;]:  Controls MPD via the " \
		"conversation window.\n" \
		"</pre");
	
	command_help[COMMAND_PLAY]   = g_strdup("Starts playing a song");
	command_help[COMMAND_PAUSE]  = g_strdup("Pauses the currently playing song");
	command_help[COMMAND_STOP]   = g_strdup("Stops the currently playing song");
	command_help[COMMAND_NEXT]   = g_strdup("Skips to the next song");
	command_help[COMMAND_PREV]   = g_strdup("Rewinds to the previous song");
	command_help[COMMAND_REPEAT] = g_strdup("Toggles repeat");
	command_help[COMMAND_RANDOM] = g_strdup("Toggles random");
	command_help[COMMAND_SEND]   = g_strdup("Sends playing song in conversation");
	command_help[COMMAND_STATUS] = g_strdup("Displays MPD status á la mpc");
	command_help[COMMAND_HELP]   = g_strdup("Displays help for this command");

	/*  Registration of a callback for when a preference is changed */
	pidgimpd_pref_change = purple_prefs_connect_callback(
		"mpd",
		"/plugins/gtk/audio/pidgimpd",
		pidgimpd_pref_change_cb,
		NULL);
	
	/*  Registration of a callback for main command */
	pidgimpd_cmd = purple_cmd_register(
		"mpc",
		"ww",
		PURPLE_CMD_P_PLUGIN,
		PURPLE_CMD_FLAG_IM |
		    PURPLE_CMD_FLAG_CHAT |
		    PURPLE_CMD_FLAG_ALLOW_WRONG_ARGS,
		NULL,
		pidgimpd_do_cmd_cb,
		help,
		NULL);
	
	/*  Register a timeout to update MPD status regularly */
	pidgimpd_update_status =
	    purple_timeout_add(700, pidgimpd_update_status_cb, NULL);

	/* Initial connect. This decided if the plugin were to be loaded before,
	 * but that makes no sense, because you cannot fix the configuration if the
	 * plugin is not loaded. ;) */
	pidgimpd_connect(NULL);
	
	return TRUE;
}

static gboolean plugin_unload(PurplePlugin *plugin)
{
	uint8_t i;

	purple_debug_misc("pidgimpd", "Unloading PidgiMPD.\n");
	
	purple_prefs_disconnect_callback(pidgimpd_pref_change);
	
	purple_cmd_unregister(pidgimpd_cmd);
	
	purple_timeout_remove(pidgimpd_update_status);
	
	for(i = 0; i < COMMAND_HELP; i++)
		g_free(command_help[i]);

	g_free(format_syntax);
	g_free(help);

	if(mpd_check_connected(mo))
	{
		mpd_free(mo);

		mo = NULL;
	}

	return TRUE;
}

static PurplePluginUiInfo prefs_info =
{
	get_plugin_pref_frame,
	0,   /* page_num (reserved) */
	NULL /* frame (reserved)    */
};


static PurplePluginInfo info =
{
	PURPLE_PLUGIN_MAGIC,        /**< magic          */
	PURPLE_MAJOR_VERSION,       /**< major version  */
	PURPLE_MINOR_VERSION,       /**< minor version  */
	PURPLE_PLUGIN_STANDARD,     /**< type           */
	NULL,                       /**< ui_requirement */ //PURPLE_GTK_PLUGIN_TYPE
	0,                          /**< flags          */
	NULL,                       /**< dependencies   */
	PURPLE_PRIORITY_DEFAULT,    /**< priority       */
	
	"pidgimpd",                 /**< id             */
	NULL,                       /**< name           */
	VERSION,                    /**< version        */
	NULL,                       /**  summary        */
	NULL,                       /**  description    */    
	NULL,                       /**< author         */
	NULL,                       /**< homepage       */

	plugin_load,                /**< load           */
	plugin_unload,              /**< unload         */
	NULL,                       /**< destroy        */

	NULL,                       /**< ui_info        */
	NULL,                       /**< extra_info     */
	&prefs_info,                /**< prefs_info     */
	NULL                        /**< actions        */
};

static void init_plugin(PurplePlugin* plugin)
{
	info.name = "PidgiMPD";
	info.summary = "Pidgin plugin for MPD.";
	info.description = "Allows for interaction with an MPD server within Pidgin.";
	info.author = "Erik Karlsson <pilo@ayeon.org>";
	info.homepage = "http://ayeon.org/projects/pidgimpd";

	purple_prefs_add_none("/plugins/gtk");
	purple_prefs_add_none("/plugins/gtk/audio");
	purple_prefs_add_none("/plugins/gtk/audio/pidgimpd");

	purple_prefs_add_string("/plugins/gtk/audio/pidgimpd/hostname", 
	    "localhost");
	purple_prefs_add_int("/plugins/gtk/audio/pidgimpd/port", 6600);
	purple_prefs_add_string("/plugins/gtk/audio/pidgimpd/password", "");
	purple_prefs_add_string("/plugins/gtk/audio/pidgimpd/songformat", 
	    "/me listens to %a - %t");
	purple_prefs_add_string("/plugins/gtk/audio/pidgimpd/showformat", 
	    "%a - %t");
	purple_prefs_add_bool("/plugins/gtk/audio/pidgimpd/title_on_away", FALSE);
	purple_prefs_add_bool("/plugins/gtk/audio/pidgimpd/send_on_new", FALSE);
	purple_prefs_add_bool("/plugins/gtk/audio/pidgimpd/show_on_new", FALSE);

	purple_prefs_add_bool("/plugins/gtk/audio/pidgimpd/pref_changed", FALSE);
}

PURPLE_INIT_PLUGIN(pidgimpd, init_plugin, info)
